#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

@author: vincent
"""

import os
import sys
import copy
import csv
import numpy as np
import time
import netCDF4 as nc
from sklearn.metrics import r2_score
from scipy.interpolate import griddata

import torch
import torch.nn as nn

from funcsinputsNN import arrayStandardize
from funcsinputsNN import regrid2Dnonan
from funcsinputsNN import pastPeriodMean3dTensor
from funcsinputsNN import pastBoundedPeriodMean3dTensor

### Set device ###
device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

   
### Functions for convolution procedures ###
def conv3x3(in_channels,out_channels,stride=1,padding=1,bias=True):
    outmodule = nn.Conv2d(in_channels=in_channels,out_channels=out_channels,
                          kernel_size=3,stride=stride,padding=padding,bias=bias)
    return(outmodule)

def upconv2x2(in_channels,out_channels,stride=2):
    outmodule = nn.ConvTranspose2d(in_channels=in_channels,out_channels=out_channels,
                                   kernel_size=2,stride=stride)
    return(outmodule)

def conv1x1(in_channels,out_channels):
    outmodule = nn.Conv2d(in_channels=in_channels,out_channels=out_channels,
                          kernel_size=1,stride=1)
    return(outmodule)

def maxpool2x2(stride=2):
    outmodule = nn.MaxPool2d(kernel_size=2,stride=stride)
    return(outmodule)

def avgpool2x2(stride=2):
    outmodule = nn.AvgPool2d(kernel_size=2,stride=stride)
    return(outmodule)

### Cropping function ###
def centeredcropping4dtensors(layerupconvol,layerdownconvol):
    '''
    Cropping the layerdownconvol around its center to match the
    dimensions of layerupconvol
    Assumes 4-dimensional tensors (Nsamples,Nchannels,nny,nnx)
        nny and nnx dimensions are matched
    '''
    upy  = layerupconvol.shape[2] #y-dim of layer in upward path (decoder)
    upx  = layerupconvol.shape[3] #x-dim of layer in upward path (decoder)
    dwy0 = layerdownconvol.shape[2] #y-dim of layer in downwward path (encoder)
    dwx0 = layerdownconvol.shape[3] #x-dim of layer in downwward path (encoder)
    if(upy<dwy0 or upy<dwy0):
        if(upy<dwy0):
            ntocrop = dwy0-upy #number of y-indices to remove
            ll      = ntocrop//2 #starting index
            tt      = dwy0-(ntocrop-ll) #ending index
            layerdownconvol = layerdownconvol[:,:,ll:tt,:]
        if(upx<dwx0):
            ntocrop = dwx0-upx #number of x-indices to remove
            ll      = ntocrop//2 #starting index
            tt      = dwx0-(ntocrop-ll) #ending index
            layerdownconvol = layerdownconvol[:,:,:,ll:tt]
    return(layerdownconvol)

### Define class for full down-convolution step ###
class DownConvBlock(nn.Module):
    '''
    Encoder
    '''
    def __init__(self,in_channels,out_channels,pooling,dropoutproba=0,stride2ndconv=1,batchnormalization=False):
        super(DownConvBlock,self).__init__()
        
        # Initialize inputs #
        self.in_channels        = in_channels
        self.out_channels       = out_channels
        self.pooling            = pooling
        self.batchnormalization = batchnormalization
        self.dropoutproba       = dropoutproba
        self.stride2ndconv      = stride2ndconv #stride to be applied in the 2nd convolution

        # First convolution #
        self.conv1   = conv3x3(self.in_channels,self.out_channels)
        self.relu1   = nn.ReLU()
        
        # Second convolution #
        if(not(self.pooling==False) and self.stride2ndconv>1):
            ValueError('stride must be 1 if we use pooling')
        self.conv2   = conv3x3(self.out_channels,self.out_channels,stride=self.stride2ndconv) #num of in channels same as out channels
        self.relu2   = nn.ReLU()
        
        # Dropout #
        if(self.dropoutproba>0):
            self.dropout = nn.Dropout2d(self.dropoutproba)
        
        # Pooling #
        if(self.pooling=='maxpool'):
            self.layerpool = maxpool2x2()
        if(self.pooling=='avgpool'):
            self.layerpool = avgpool2x2()
            
    def forward(self,xx):
        # First convolution #
        xx = self.conv1(xx)
        xx = self.relu1(xx)
        # If stride=2 for 2nd conv: need to save prepool here #
        if(self.stride2ndconv==2):
            prepoolxx = torch.clone(xx)
        # Second convolution #
        xx = self.conv2(xx)
        xx = self.relu2(xx)
        # Batch normalization #
        if(self.batchnormalization):
            xx = nn.functional.batch_norm(xx,running_mean=None,running_var=None,training=True)
        # Save before pooling for concatenation #
        if(self.stride2ndconv==1):
            prepoolxx = torch.clone(xx)
        # Dropout #
        if(self.dropoutproba>0):
            xx = self.dropout(xx)
        # Pooling #
        if(not(self.pooling==False)):
            xx = self.layerpool(xx)
        return(xx,prepoolxx)

### Define class for full up-convolution step ###        
class UpConvBlock(nn.Module):
    '''
    Decoder
    '''
    def __init__(self,in_channels,out_channels,skipconnection=True,batchnormalization=False):
        super(UpConvBlock,self).__init__()
        
        # Initialize inputs #
        self.in_channels        = in_channels
        self.out_channels       = out_channels
        self.skipconnection     = skipconnection
        self.batchnormalization = batchnormalization
        
        # Up-convolution #
        self.upconv   = upconv2x2(self.in_channels,self.out_channels)
        
        # First convolution #
        if(self.skipconnection):
            numinchconv   = 2*self.out_channels #concatenating feature map from the down-sampling path
        else:
            numinchconv   = self.out_channels #no concatenation of feature map from the down-sampling path
        self.conv1    = conv3x3(numinchconv,self.out_channels) 
        self.relu1    = nn.ReLU()
        
        # Second convolution #
        self.conv2    = conv3x3(self.out_channels,self.out_channels) #num of in channels same as out channels
        self.relu2    = nn.ReLU()
        
    def forward(self,xxfromdownpath,xxfromuppath):
        
        # Up-convolution #
        fromup = self.upconv(xxfromuppath)
        
        if(self.skipconnection):
            # Crop the layer from the encoder (fromup) if necessary #
            fromdown = centeredcropping4dtensors(fromup,xxfromdownpath)
            # Concatenation #
            xx = torch.cat((fromup,fromdown),dim=1)
        else:
            xx = torch.clone(fromup)
        
        # First convolution #
        xx = self.conv1(xx)
        xx = self.relu1(xx)
        
        # Second convolution #
        xx = self.conv2(xx)
        xx = self.relu2(xx)
        
        # Batch normalization #
        if(self.batchnormalization):
            xx = nn.functional.batch_norm(xx,running_mean=None,running_var=None,training=True)
        
        return(xx)

### Define Convolutional Neural Network class ###
class U_Net(nn.Module):
    '''
    U-Net architecture for Convolutional Neural Network
    Based on Ronneberger et al. (2015)
    '''
    def __init__(self,numinchannels,numoutchannels,depth,pooling,numfiltersinit,skipconnections,numscal,dropoutprobas,batchnormalization=False):
        super(U_Net,self).__init__()
        
        # Initialize inputs #
        self.numinchannels   = numinchannels
        self.numoutchannels  = numoutchannels
        self.depth           = depth
        self.numfiltersinit  = numfiltersinit #number of initial filters
        self.numscal         = numscal #number of scalar values concatenated at the end of the encoding stage
        self.skipconnections = skipconnections #whether skip connections are allowed at the different depth levels
        if(len(skipconnections)!=self.depth-1):
            ValueError('skipconnections should be a list of length depth-1')
        self.dropoutprobas = dropoutprobas #dropout proba applied at end of each down-convolution block
        if(len(dropoutprobas)!=self.depth):
            ValueError('dropoutprobas should be a list of length depth')
        if(type(pooling)==str or pooling==False):
            self.pooling = [pooling for ii in range(depth-1)]+[False]
        elif(type(pooling)==list and len(pooling)==depth-1):
            self.pooling = pooling+[False]
        elif(type(pooling)==list and len(pooling)==depth):
            self.pooling = pooling
            if(pooling[-1]!=False):
                ValueError('pooling last argument must be False if of length depth')
        else:
            ValueError('pooling should be a list of length depth-1')
        if(np.array(pooling).dtype=='bool' and np.all(np.array(pooling)==False)):
            print('   No pooling, using 2-stride down convolutions')
        
        # Prepare lists of down and up convolution modules #
        self.downConvs = []
        self.upConvs   = []
        
        # Encoder #
        previousnumout = 0
        for ii in range(self.depth):
            if(ii==0):
                iinumin  = copy.deepcopy(self.numinchannels) #numinput channels
            else:
                iinumin  = copy.deepcopy(previousnumout) #input channels are output channels of last conv
            iinumout = self.numfiltersinit*(2**ii) #double number of feature maps at each block
            if(ii==self.depth-1):
                iinumout = iinumout-self.numscal #leave space for concatenating the scalValues
            if(self.pooling[ii]==False and ii<self.depth-1):
                stride2ndconv = 2
            else:
                stride2ndconv = 1
            iidownconv = DownConvBlock(iinumin,iinumout,pooling=self.pooling[ii],dropoutproba=self.dropoutprobas[ii],
                                       stride2ndconv=stride2ndconv,batchnormalization=batchnormalization)
            self.downConvs.append(iidownconv)
            previousnumout = copy.deepcopy(iinumout)
        
        # Decoder (over depth-1 blocks) #
        for jj in range(self.depth-1):
            if(jj==0):
                previousnumout += self.numscal #we concatenated the scalValues
            jjnumin  = copy.deepcopy(previousnumout) #input channels are output channels of last conv
            jjnumout = jjnumin//2 #up-sampling by 2 at each up-convolution
            if(jj<self.depth-2):
                jjupconv = UpConvBlock(jjnumin,jjnumout,
                                       skipconnection=self.skipconnections[-1-jj],batchnormalization=batchnormalization) #optional batchnormalization
            elif(jj==self.depth-2):
                jjupconv = UpConvBlock(jjnumin,jjnumout,
                                       skipconnection=self.skipconnections[-1-jj],batchnormalization=False) #no batchnormalization
            self.upConvs.append(jjupconv)
            previousnumout = copy.deepcopy(jjnumout)
            
        # Final convolution layer #
        self.finalConv = conv1x1(previousnumout,self.numoutchannels)
        
        # Convert module lists #
        self.downConvs = nn.ModuleList(self.downConvs)
        self.upConvs   = nn.ModuleList(self.upConvs)
        
        # Initialize parameters of U-Net #
        self.initialize_parameters(wInitMethod=nn.init.kaiming_normal_)
        print('U-Net model fully intialized')
        
    @staticmethod
    def weightinitialization(module,initmethod):
        if(isinstance(module,(nn.Conv2d,nn.ConvTranspose2d))):
            initmethod(module.weight) #initialize weights of module according to method
    
    @staticmethod
    def biasinitialization(module,initmethod):
        if(isinstance(module,(nn.Conv2d,nn.ConvTranspose2d))):
            initmethod(module.bias) #initialize bias of module according to method
    
    def initialize_parameters(self,wInitMethod=nn.init.kaiming_uniform_,
                                   bInitMethod=nn.init.zeros_):
        '''
        U-Net parameter initialization
        Methods for weights and biases can be chosen separately
        Note: Kaiming Uniform is equivalent to He method
        '''
        for module in self.modules():
            self.weightinitialization(module,wInitMethod)
            self.biasinitialization(module,bInitMethod)
    
    def forward(self,xx,scalVals):
        
        # Save encoder outputs at each step #
        outputEncoder = []
        
        # Down-convolution pathway #
        for ii,module in enumerate(self.downConvs):
            xx,prepoolxx = module(xx)
            outputEncoder.append(prepoolxx)
        
        # Discard last prepoolxx as no pooling was performed on last step #
        outputEncoder = outputEncoder[0:-1]
        
        # Concatenate scalar values as a feature map to xx #
        if(scalVals.shape[1]!=self.numscal):
            ValueError('Dim[1] of scal values passed must match numscal')

        scalVals = scalVals.unsqueeze(dim=2).unsqueeze(dim=3)
        scalVals = scalVals.repeat(1,1,xx.shape[2],xx.shape[3])
        xx = torch.cat((xx,scalVals),dim=1)
        
        # Up-convolution pathway #
        for jj,module in enumerate(self.upConvs):
            toconcat = outputEncoder[-(jj+1)]
            xx       = module(toconcat,xx)
       
        # Final 1x1 convolution #
        xx = self.finalConv(xx)
        return(xx)
    



